/*     
    v0.4 (c) JS 2001
       
    make D=1 lattices from one unitcell
       
    example:
       
     5  ------0---1----5---6----
     4         \ /      \ /
     3          2        7
     2         / \      / \    
     1  ------3---4----8---9----
     0   
             0123456
             
  v0.3 - J1J2-chain added, p_t set
                       
*/
#include <stdlib.h>
#include <stdio.h>

#define PChain
#undef J1J2Chain
#undef KagomeChain
#undef TriangleChain
#undef DiamondChain

/* S=1-S=1/2-P-chain */
#ifdef PChain
#define NUMsites 4
#define NUMbonds 9 
#define NUMidx   2 
int evec[2]={3,99};
int sites[NUMsites][2]={{1,1},{1,2},{2,1},{2,2}};
int bonds[NUMbonds][4]={ /* site1 site2 idx-J cell-x */
 {0,2,1,0}, {0,3,1,0}, {1,2,1,0}, {1,3,1,0},
 {2,0,1,1}, {2,1,1,1}, {3,0,1,1}, {3,1,1,1},
 {0,1,2,0}};
#endif


/* J2=2J1 */
#ifdef J1J2Chain
#define NUMsites 1
#define NUMbonds 2
#define NUMidx   2
int evec[2]={1,99};
int sites[NUMsites][2]={{1,1}};
int bonds[NUMbonds][4]={ /* site1 site2 idx-J cell-x */
 {0,0,1,1},  {0,0,2,2}};
#endif

/* J2=(3/2)J1 */
#ifdef KagomeChain
#define NUMsites 5
#define NUMbonds 8
#define NUMidx   2
int evec[2]={4,99};
int sites[NUMsites][2]={{1,3},{3,3},{2,2},{1,1},{3,1}};
int bonds[NUMbonds][4]={ /* site1 site2 idx-J cell-x */
 {0,1,1,0},  {3,4,1,0}, {0,2,1,0},  {2,4,1,0},
 {1,2,1,0},  {2,3,1,0}, {1,0,2,1},  {4,3,2,1}};
#endif

/* J2=2J1 */
#ifdef TriangleChain
#define NUMsites 2
#define NUMbonds 3
#define NUMidx   2
int evec[2]={2,99};
int sites[NUMsites][2]={{1,1},{2,2}};
int bonds[NUMbonds][4]={ /* site1 site2 idx-J cell-x */
 {0,1,2,0},  {1,0,2,1}, {0,0,1,1}};
#endif

#ifdef DiamondChain
#define NUMsites 3
#define NUMbonds 6
#define NUMidx   1
int evec[2]={2,99};
int sites[NUMsites][2]={{1,1},{1,3},{2,2}};
int bonds[NUMbonds][4]={ /* site1 site2 idx-J cell-x */
 {0,2,1,0},  {1,2,1,0}, {0,0,1,1},
 {1,1,1,1},  {2,0,1,1}, {2,1,1,1}};
#endif

int initvars(int n0,int pbc){
   int i,j,nn,nw;
   nn=n0*NUMsites;
   nw=n0*NUMbonds;
   printf("# def-file, 1D-chain\n"
	"# automaticly generated by m_1D v0.1 n0=%d PBC\n"
	"#===edge=== %d 0 0 999\n"
	"#+++model+++\n"
	"# NN=Number_of_sites Nw=Number_of_bonds Nv=Number_of_bonds(2)\n"
	"%d %d 0\n"
	"# s  X  Y  Z p_e p_U (s=siteindex  (X Y Z)=position  p=index of parameter)\n",
     n0,n0*evec[0],nn,nw);
   for(i=0;i<n0;i++)for(j=0;j<NUMsites;j++)
     printf(" %2d %2d %2d 0 8 7\n",
        i*NUMsites+j, sites[j][0]+i*evec[0], sites[j][1]);
   printf("# Bond s1 s2 p_t p_J p_V p_w\n");
   for(i=0;i<n0;i++)for(j=0;j<NUMbonds;j++)
     printf(" %2d %2d %2d  %d %c%d 0 %d\n",
      i*NUMbonds+j,
      bonds[j][0]+NUMsites*i,
      bonds[j][1]+NUMsites*((i+bonds[j][3]+n0)%n0),
      bonds[j][2]+NUMidx,
      (((i+bonds[j][3]>=n0 || i+bonds[j][3]<0) && pbc==1)?'+':' '),
      bonds[j][2],
      bonds[j][2]+2*NUMidx+1);
   printf("#---model---\n");
   return 0;
}


main(int argc,char *argv[]){
   if(argc!=2){
     fprintf(stderr,
       " generate daten.def for an 1D chain v0.1 JS01\n"
       " use: m_1D ?length?\n"
       "   length: 1..6  chain length in unit cells\n");
     exit(1);
   }
   initvars(atoi(argv[1]),1);
}
