/*
   history:
    Mar99 - lanz2 gegen ueberlauf geschuetzt, (starke Aenderung des 
            vektors bei unguenstigen Startbedingungen)
*/

#include "mem.h"
#include "spinstri.h"
#include "spins.h"
#include "block.h"
#include "vector.h"
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif

zahl ta[MAX_ITR];
zahl tb[MAX_ITR];		/* a,b Tridiagonalmatrix		*/
zahl tc[MAX_ITR][NEV];		/* Eigenvektoren der Diagonalmatrix	*/
    
/* --- --- Berechnung des kleinsten EW mit Lanczos (Gagliano) ----------   
  - Projektionsvektor |Pi> (Startvector |P0>)
  - zweiter Ebenenvektor |Pis>=((H-<H>_i)|Pi>)/(<H^2>_i-<H>_i^2)^(1/2)
    senkrecht zu |Pi> (<Pi|Pis>=0)
  - berechne 2x2-Matrix f"ur |Pi> und |Pis> und deren EW
  - |Pi+1>=a|Pi>+b|Pis> ist normierter EV zu kleinerem EW

  v0=|Pi> v1=H|Pi> v2=HH|Pi>
  h=<H>_i hh=<HH>_i hhh=<HHH>_i
  b=q(<H^2>_i-<H>_i^2)=q(hh-h^2)          mit  q(x)=x^(1/2)
  H-Matrix=[[<Pi|H|Pi>,<Pi|H|Pis>],[<Pis|H|Pi>,<Pis|H|Pis>]]
          =[[h,b],[b,(hhh-2*hh*h+h^3)/b^2]]
          =[[h,b],[b,(hhh-h*hh)/b^2-h]]
          =[[h,b],[b,2*b*f+h]]
  EW=(a11+a22)/2-q((a11+a22)^2/4-a11*a22+a12^2)
    =(hhh-h*hh)/(2b^2)-q((hhh-h*hh)^2/(4b^4)-h*((hhh-h*hh)/b^2-h)+b^2)
    =(hhh-h*hh)/(2b^2)-q((hhh-h*hh)^2/(4b^4)-h*(hhh-h*hh)/b^2+hh)
    =h+b*f-q((b*f+h)^2-2*h*b*f-h^2+b^2) ,    a22=2*b*f+h
    =h+b*f-q(b^2*f^2+b^2)
    =h+b*(f-q(1+f*f))                   ,    alf=f-q(1+f*f)
    =h+b*alf
  EV=[(EW-a22)/a12,1]/q((EW-a22)^2/a12^2+1) * [Pi,Pis]
    =((EW-a22)/b*v0+(1/b*v1-h/b*v0))/q((EW-a22)^2/b^2+1)
    =((EW-a22-h)*v0+v1)/q((EW-a22)^2+b^2)
  f=(a22-a11)/(2*a12),     -p/2=(a11+a22)/2=a22-a12*f=a12*f+a11=b*f+h
   --- last change 7.12.98 JS ueberarbeitet ---
 * --- --- -------------------------------------------------- ---- --- */
/* --- Inp: v0  Out: v0,itr,ew  Wrk: v1 */
int lanz1(tvec *v0,tvec *v1,tvec *ev){
    tvec *v2; cplx a,alf,ak,b,f,r,h,hh,hhh;
    int itr,j,blk; ULong i; double eps=1e-6;	
    /* eps=precission^2 = <(H-<H>)^2> = <HH>-<H><H> >= 0, also complex ??? */
    v2=ev;		/* new 12.09.1998 use memory of ev[nq] */
    r=ak=alf=f=0;
    for(itr=0;;itr++){
      clrvec(v1);hamilton(v0,v1);		/* v1=H*v0+v1 */
      clrvec(v2);hamilton(v1,v2);
      for(hhh=hh=h=0,blk=0;blk<B_NUM;blk++)for(i=0l;i<b_len[blk];i++){
        r=Conj(M_f2d(v0->v[blk][i]));
        h += r*M_f2d(v1->v[blk][i]);
        hh+= r*M_f2d(v2->v[blk][i]);
        hhh+=Conj(M_f2d(v1->v[blk][i]))*M_f2d(v2->v[blk][i]); }
      b=hh-h*h;		/* negativ if rounding-errors! cmplx>0 ??? */
      b=fabs(b);	/* if cmplx (gegen 0) */
      if(D(b)>0){	/* neuer Projektionsvektor senkrecht zum alten */
        b=sqrt(b); 				/* 2x2 diagonalization */
        f=(hhh-3*h*hh+2*h*h*h)/(2*b*b*b);	/* gegen inf */
        alf=f-sqrt(1.+f*f);			/* gegen -0 bis f*f OVL */ 
        a=1./sqrt(1.+alf*alf);			/* gegen  1 */
        ak=a*alf/b;				/* gegen -0 */
        r=a-ak*h;				/* gegen  1 */
        for(blk=0;blk<B_NUM;blk++)for(i=0l;i<b_len[blk];i++)
        v0->v[blk][i]=M_d2f(r*M_f2d(v0->v[blk][i])+ak*M_f2d(v1->v[blk][i]));
        /* h=h+b*alf; */ }
      printf("\n i=%3d h=%16.12f alf=%.4e b=%.1e v=[% .3f,% .3f] t=%s",
             itr,D(h),D(alf),D(b),D(r),D(h*ak),sclock(~0l));
      if(/* itr>1 && */ fabs(b)<=eps) break; }
    for(blk=0;blk<B_NUM;blk++)for(i=0l;i<b_len[blk];i++)
      ev->v[blk][i]=v0->v[blk][i];
    for(j=0;j<NEW;j++)ew[j]=D(h);
    return(itr); }
      
/* ----------- bisection method -------------------------------------
 * calculate the eigenvalues of a tridiagonal matrix T=(tb,ta,tb)
 * --- Inp: ta,tb=Tridiagonalmatrix(T),  nn=size of the matrix T   ---
 * ---      ne=number of eigenvalues ew to calculate, ep=precision  ---
 * --- Out: ew  eigenvalues of T via "Nullstellen" of Det(A-x*I) --- */
/* --- was passiert bei Entartung? => EW nur einfach! --- */
void bisec(int nn,int ne,zahl ep){
    static zahl b2[MAX_ITR]; /* in DOS not enough stackspace, faster */
    zahl ra,rb,rc,g,epa,range; int i,j,k,neg,ipass;
    range=fabs(ta[0])+fabs(tb[0]);
    for(k=1;k<nn-1;k++) range=max(range,fabs(tb[k-1])+fabs(ta[k])+fabs(tb[k]));
    range=max(range,fabs(tb[nn-2])+fabs(ta[nn-1]));range=-range;
    b2[0]=0;for(i=1;i<nn;i++)b2[i]=Norm(tb[i-1]);
    epa=fabs(range*ep);for(i=0;i<ne;i++)ew[i]=-range;rb=range; /* rb...ra */
    for(k=0;k<ne;k++){ ra=ew[k];	/* ra=upper bound, rb=lower bound */
      /* was j<100, but this is a bit to small ??? */
      for(j=0;j<10000;j++){		/* sturm sequencing */
        if(fabs(ra-rb)<epa)break;	/* ra=rb => break */
        rc=(ra+rb)*0.5; ipass=neg=0; g=1.0;
        for(i=0;i<nn;i++){   /* g=Det_{i}(rc)/Det_{i-1}(rc) */
          if(ipass==0)g=rc-ta[i]-b2[i]/g;else	/* do not divide if g=0 */
          if(ipass==1)ipass=2;
          else      { ipass=0; g=rc-ta[i]; }	/* new start, decoupled */
          if(ipass==0){ if(g<0.0)neg++;		/* sign-change of Det */
            if(fabs(g)<=fabs(b2[i]*epa*ep)) ipass=1; } /* g==0 ? */
        }   
        neg=nn-neg;   /* anzahl nullstellen rc...ra -> rb...rc */
        if(neg<=k)rb=rc;		/* new lower bound */
        else    { ra=rc; 		/* new upper bound */
                  for(i=k;i<min(neg,ne);i++)ew[i]=rc; }
      } }
    return; }
/* --- eigenvectors (EV) of a tridiagonal matrix by inverse iteration */
/*  see Numerical Recipes  (Feb99)
 *   A - matrix
 *   a - start-eigenvalue
 *   x - startvector (random)
 *   step1: solve equation system (A-a*1)*y==x for vector y (ex: LU decomposition)
 *   step2: x := y/|y|
 *   step3: a := a+(x,x)/(x,y)  ( not necessary every step )
 *   if Ax=ax then break else goto step1   
 */
/* --- nn i Matrixgroesse, {ta,tb} TriDiagMatrix -> tc EVs */
/* --- fuer Entartung muss EW in ew[] 2mal enthalten sein! */
zahl di[MAX_ITR],bl[MAX_ITR],bu[MAX_ITR],
     bv[MAX_ITR],cm[MAX_ITR];
char lex[MAX_ITR];	/* long to char 1998-182 */
void vec12(int nn,int nev){
    zahl s,prd;int i,j,k,km,l;
    STKa("vec12-lanczos",(unsigned long *)&j);
    km=0;/* km=0 added by JS */
    for(k=0;k<nev;k++){
        for(j=0;j<nn;j++){ di[j]=ew[k+NE0]-ta[j];bl[j]=bu[j]=-tb[j];}
        /* --- LU decomposition */ 
        for(j=0;j<nn-1;j++){
           if(fabs(di[j])>fabs(bl[j])){		/* --- non pivoting */
              lex[j]=0;if(fabs(di[j])<1e-30)di[j]=1e-30; /* e-40 => e-30 (float) */
              cm[j+1]=bl[j]/di[j]; di[j+1]-=cm[j+1]*bu[j]; bv[j]=0; }  
           else {				/* --- pivoting */
              lex[j]=1;cm[j+1]=di[j]/bl[j];di[j]=bl[j];s=bu[j];
              bu[j]=di[j+1];bv[j]=bu[j+1];di[j+1]=s-cm[j+1]*bu[j];
              bu[j+1]=-cm[j+1]*bv[j]; }
        }  
        if(fabs(di[nn-1])<1e-30)di[nn-1]=1e-30;
        /* --- initial vec */ 
        for(j=0;j<nn;j++)tc[j][k]=1./((j+1)*5.); /* random ??? */
        /* --- degeneracy check up -> Entartung km..k */
        if(k==0) km=k; else 
        if(fabs(ew[k+NE0]-ew[km+NE0])>1e-10) km=k;  /* frueher 1e-13 1998-184 */ 
        else for(i=km;i<k;i++){		/* orthogonalise ini_vec to tri_ev */
            for(prd=0,j=0;j<nn;j++) prd+=tc[j][i]*tc[j][k];  /* <c_i|c_k> */
            for(      j=0;j<nn;j++) tc[j][k]-=prd*tc[j][i]; }
        /* --- inverse iteration */
        for(l=0;l<(k-km+3);l++){	/* was soll l machen, wozu ??? */
            if((l!=0)||(k!=km)){ /* --- forward substitution */
               for(j=0;j<nn-1;j++)
               if(lex[j]==0) tc[j+1][k]-=cm[j+1]*tc[j][k];
               else {s=tc[j][k];tc[j][k]=tc[j+1][k];
                     tc[j+1][k]=s-cm[j+1]*tc[j][k];}
            }  /* --- backward substitution */
            for(j=nn-1;j>=0;j--){
                s=tc[j][k];
                if(j<(nn-1))s-=bu[j]*tc[j+1][k];
                if(j<(nn-2))s-=bv[j]*tc[j+2][k];
                tc[j][k]=s/di[j];
            }   /* --- normalization */
            for(prd=0,j=0;j<nn;j++) prd+=Norm(tc[j][k]);
            if(prd>1e-30)prd=1./sqrt(prd);else prd=0;
            for(j=0;j<nn;j++) tc[j][k]*=prd;
        }
    }   STKe();return;
}

int konvergenz_i=0;		/* set to 0 for start, Mai99 JS */
#if 1
/* new version hopefully work better Mai99 JS */
int konvergenz(zahl *ew,zahl *et,zahl eps){	/* from inc1() */
  int r=0,k,in; static double maxew=0.0;
  in=min(max(1,pew),NEW);
  if(!konvergenz_i){ maxew=ew[in-1];konvergenz_i=1; }
  if( ew[in-1]<maxew )
  if(fabs(et[0]-ew[0])<=fabs(eps*ew[0]))
  for(k=0;k<in;k++)
    if(fabs(et[k]-ew[in-1])<=fabs(eps*ew[k])){ r=1; break; }
  for(k=0;k<NEW;k++)et[k]=ew[k];
  return r;}
#else
/* old version < Mai99, often too much steps */
int konvergenz(zahl *ew,zahl *et,zahl eps){	/* from inc1() */
  int r,k; r=1;
  for(k=0;k<min(max(1,pew),NEW);k++)
    if(fabs(et[k]-ew[k])>fabs(eps*ew[k])){ r=0; break; }
  for(k=0;k<NEW;k++)et[k]=ew[k];
  return r;}
#endif

/* --- --- Berechnung der kleinsten EW mit Lanczos --------------------- */
/* --- Inp: iv  Out: a,b,itr,ew  Wrk: v0,v1 ret: Iterationen */
int inc1(int nev,tvec *v0,tvec *v1){
    ULong j;int i,k,blk;zahl a1,b1,eps,et[NEW],rb;cplx t1; tvec *vtmp;
    STKa("inc1-lanczos",(unsigned long *)&i);
    if(vvv)printf(" inc1-lanczos");clrvec(v0);konvergenz_i=0;
    for(k=0;k<NEW;k++)ew[k]=et[k]=13.5;
#if Zahl & 1
    eps=1e-11;/* high precision 1e-11 */
#else
    eps=1e-9;/* low precision */
#endif
    if( checkpoint<4 )
    {
      for(i=1;i<=MAX_ITR;i++){		/* Iteration bei 0 beginnen !!! */
       /* Normierung gleich einbauen! v1[]*=r1;v0[]*=r0 */
       /* t=<v1*H*v1>,v0=H*v1+v0, <v1|v0>==0 !def! */
       t1=hamilton(v1,v0);
       ta[i-1]=a1=Re(t1);b1=0;	/* a1=<v1*H*v1>,v0=H*v1+v0 */
       for(blk=0;blk<B_NUM;blk++)for(j=0l;j<b_len[blk];j++){ /* call mlply_av ret_Norm ? */
         v0->v[blk][j]-=M_d2f(a1*M_f2d(v1->v[blk][j]));
         b1+=Norm(M_f2d(v0->v[blk][j]));
       }
       tb[i-1]=b1=sqrt(b1);
       /* or store normalization factor seperately ? */
       if(b1>0)for(rb=1/b1,blk=0;blk<B_NUM;blk++)for(j=0;j<b_len[blk];j++){
         v0->v[blk][j]=M_d2f( rb*M_f2d(v0->v[blk][j]));
         v1->v[blk][j]=M_d2f(-b1*M_f2d(v1->v[blk][j])); }
       vtmp=v0;v0=v1;v1=vtmp; /* faster change using pointer */
       outtri(i,a1,b1);
       if( sig1==2 ) { /* save before possible crash */
         savevec(v0,"tmp/v0.dat",-1,-1);
         savevec(v1,"tmp/v1.dat",-1,-1);printf(" vx saved, it=%d",i);
       }
       IFcplx(if(fabs(t1.imag())>1e-6)E1("Im<H>=% .0e",D(t1.imag()));) /* flt=1e-6 */
       if((i>19 && (i%5)==0) || (ULong)i+5>n1 || b1<1e-15){ /* Konvergenztest */
          /* calculate the eigenvalues of the tridiagonalmatrix T */
          bisec(i,NEW,1e-13);nl(stdout);outew(stdout,i);
          if(b1<1e-15){printf(" inc1_b<1e-15");break;}
          if(konvergenz(ew,et,eps))break; if((ULong)i>n1)break; /* Mai99 JS */
          if(i>=MAX_ITR){printf(" keine Konvergenz");break;}
       }
      }
      printf(" checkpoint=%d",++checkpoint);
    } else { /* reload old values */
      i=loadtri(ta,tb,MAX_ITR);
      bisec(i,NEW,1e-13);nl(stdout);outew(stdout,i);
    }
    /* ToDo: vec12 should be called by incv1, because output is needed only there */
    if(nev)vec12(i,nev);STKe();return(i);}
    
/* --- eigenvector by the Lanczos method (after vec12())
 * --- convert eigenvectors of tridiagonal matrix back to H space 
 * --- Inp: a,b,c --- Out: v0,v1,ev=linear_access(3+itr) */
void incv1(int nev,tvec *v0,tvec *v1,tvec ev[],int itr){
    ULong j,ii;int i,k,blk;zahl a1,b1,rb,rc; tvec *vtmp;
    if(nev<1) return;ii=1l;
    STKa("incv1-lanczos",(unsigned long *)&ii);
    clrvec(v0);
    for(k=0;k<nev;k++)
    for(blk=0;blk<B_NUM;blk++)for(j=0l;j<b_len[blk];j++)
      ev[k].v[blk][j]=M_d2f(tc[0][k]*M_f2d(v1->v[blk][j]));
    for(i=1;i<itr;i++,ii++){ /* v1.7.8 99% time spend for hamilton() */
       hamilton(v1,v0); /* v0=H*v1+v0, ret=<v1|H*v1> faster if v0=r1*H*v1+r2*v0 ???  */
       a1=ta[i-1];b1=-tb[i-1];rb=-1.0/b1; rc=tc[i][0];
       for(blk=0;blk<B_NUM;blk++)for(j=0l;j<b_len[blk];j++){
          v0->v[  blk][j] =M_d2f(  (M_f2d(v0->v[blk][j])
                                -a1*M_f2d(v1->v[blk][j]))*rb );
          ev[0].v[blk][j]+=M_d2f(rc*M_f2d(v0->v[blk][j]));
          v1->v[blk][j]=M_d2f(M_f2d(v1->v[blk][j])*b1);
       }
       for(k=1;k<nev;k++){ rc=tc[i][k]; /* divide k=0 and k>0 => faster */
         for(blk=0;blk<B_NUM;blk++)for(j=0l;j<b_len[blk];j++)
         ev[k].v[blk][j]+=M_d2f(rc*M_f2d(v0->v[blk][j]));
       }
       vtmp=v0;v0=v1;v1=vtmp; /* faster change using pointer */
    }   
    for(k=0;k<nev;k++)normvec(&ev[k]);
    if(vvv&1){printf(" incv1:%s",sclock(~0l));}STKe();return;}
/* ------------------------------------------------------------- */

/* --- all what be needed for pthreads of diag2 --- */
typedef struct sp_diag2 {
#ifdef CONFIG_PTHREAD
  pthread_t pt;
#endif
  thbuf *hbuf;
  int *ic;
  double vv,e1,vhv,vv2,vhv2,krit;
} tp_diag2;

/*
    void-function for pthreads 
    
    at start vhv is estimated
    store vhv_old=vhv, calculate vhv_last and vhv=vhv_new
    corrected vhv = (vhv_new-vhv_old)+(vhv_new-vhv_last) ???
 */
void diag2_thread( void *ptr ){
    tp_diag2 *p=(tp_diag2 *)ptr;
    tvec *v0; mcplx vnew,vold;
    cplx b,c,nv,nhv,b1,r;
    double e1,r1,nhn,vhv,vv,krit=0;
    int ic,blk; ULong n;
    blk=p->hbuf->block;
    e1 =p->e1;
    vhv=p->vhv;
    vv =p->vv;
    ic =*(p->ic);
    v0 =p->hbuf->v0;
    for(n=b_ofs[blk];n<b_ofs[blk]+b_len[blk];n++)
    {    /* like MC mit festem walk? */
       vold=v0->v[blk][n-b_ofs[blk]];      /* float, double ... */
       nv=M_f2d(vold);                     /* <nv> n-te Komponente von |v> */
#if 0
       nhn=D(hamilton_nhv(n,NULL,p->hbuf,1)); /* <n|Hn> Diagonal-H von |n> */
       nhv=hamilton_nhv(n,v0,p->hbuf,0); /* <nHv> n-te Komponente von H|v> */
#else
       			      /* <n|Hn> Diagonal-H von |n>      */
       nhv=fast_nhv(n,v0,&nhn,p->hbuf); /* <nHv> n-te Komponente von H|v> */
#endif
       if(ic==0){ vhv+=D(Conj(nv)*nhv); continue; } /* only calculate vhv_last */
       /* ----------------------- Matrixel. von S^(-1)H*(1-nv*nv)  */
       b=vv*nhv-vhv*nv; /* a=vv*nhn-nv*nhv; */ if(fabs(b)<1e-30)continue;
       c=   nhv-nhn*nv; /* d=   vhv-nhv*nv; */
       /* <b|c>==vv*<nhv|nhv>+vhv*nhn*<nv|nv>-vv*nhn*<nhv|nv>-vhv*<nv|nhv> ??? */
       r1=vhv-vv*nhn;   /* d-a */
       if(fabs(b*c)<1e-12){continue;}  /* no changes in v0 */
       r =r1-sqrt(r1*r1+4*b*c);	/* complex??? ew r1^2-4bc=(a+d)^2-ad-bc=(p/2)^2-q */
       b1=2*b/r;                /* correction b1 is small??? */
       vnew=vold+M_d2f(b1);     /* rounding error ??? vold==nv */
       b1=M_f2d(vnew-vold);     /* set b1 according to rounding error */
       v0->v[blk][n-b_ofs[blk]]=vnew;
       krit+=Norm(b1);	/* Norm(old.v-new.v), better use max(fabs(b1))??? */
       /* vv =<v+b|v+b>  =<v|v>  +2<b|nv>  +<b|b> */
       /* vhv=<v+b|H|v+b>=<v|H|v>+2<bn|Hv>+<b|H|b> */
       /* not ok! complex ????????? */
       vv  +=D(2*Conj(b1)*nv )+Norm(b1);     /* korrekt besides matherrors */ 
       vhv +=D(2*Conj(b1)*nhv)+Norm(b1)*nhn;	/* new EW allways lower! */
       /* bad startvec can lead to overflows or big_summed_error */
       if( vv<0.5 || vv>1.6 || fabs(vhv/vv-e1)>4 ){
         printf(" warning: dE=%e vv=%f b1=%f\n",fabs(vhv/vv-e1),vv,D(b1));
         /* *(p->ic)=-1; */ 
         break;
       }
    } /* for(n) */
    p->krit=krit;
    p->vhv2=vhv;
    p->vv2 =vv;
#ifdef CONFIG_PTHREAD
    pthread_exit(0);
#endif
}

/* --- --- Berechnung des kleinsten EW mit special-Lanczos ----------
 *   see 2x2.tex for explanations 
 * --- Inp: iv,v0=startvektor  Out: itr,ew,v0=eigenvektor  ---------- */
int lanz2(tvec *v0){
    double r,e1,vhv,vv,eps=1e-9,krit;
    int itr,ic,blk,j; ULong /* i, */ ti;
    tp_diag2 pv[B_NUM],*p;
    pv[0].hbuf=alloc_hbuf();
#ifdef CONFIG_PTHREAD
    for(ic=1;ic<B_NUM;ic++)pv[ic].hbuf=alloc_hbuf();
#else
    for(ic=1;ic<B_NUM;ic++)pv[ic].hbuf=pv[0].hbuf;
#endif
#if (Zahl & 3) == 0
    eps=1e-8;  		/* float JS-Apr99 */
#endif        
#if (Zahl & 8)
    eps=1e-4;  		/* float16 JS-Feb02 find more elegant way! */
#endif        
    printf(" lanz2: eps=%5.0e\n",eps);
    vhv=0.0; e1=13.5; vv=1.0; ti=itime();
    for(itr=1,ic=0;itr<4*MAX_ITR;itr++,ic=(ic+1)&31){
       /* ToDo: reset ic on external signal sig1==1 ??? */
       krit=0;
       if (ic==0) { vhv=0; krit=999; }
       for(blk=0;blk<B_NUM;blk++){ p=&(pv[blk]);
         p->hbuf->v0   =v0;
         p->hbuf->block=blk;
         p->vv         =vv;
         p->vhv        =vhv;
         p->e1         =e1;
         p->ic         =&ic;
#ifdef CONFIG_PTHREAD
         if( pthread_create((pthread_t *)&(p->pt),
                       (const pthread_attr_t *)NULL, 
                       (void * (*)(void *)) &diag2_thread,
                       (void *) p) !=0 )  F0("pthread_create");
#else
         diag2_thread( p );
#endif
       }       
       for(blk=0;blk<B_NUM;blk++){ p=&(pv[blk]);
#ifdef CONFIG_PTHREAD
         pthread_join(p->pt,NULL);
#endif
         vv  += p->vv2  - p->vv;
         vhv += p->vhv2 - p->vhv;
         krit+= p->krit;
       }
       r=1/sqrt(vv);  mltply_av(r,v0); /* v0*=r */
       krit*=fabs(r); krit=sqrt(krit/(double)n1);
       e1=vhv/=vv; vv=1.0;		/* normieren */
       /* if(!(itr%5)) */ {
         printf("%3d %20.15f %5.0e %4s %2d %6.0e\n",
         itr,D(e1),(double)krit,sclock(~0l),ic,1-r);
         if(krit<eps)break;}
       if(itime()-ti>60*60*12l){ ti=itime(); /* all 12h -> save v0 */
         /* savevec(v0,"d1.vec",-1,-1); # better msync? */ }
    }
    for(j=0;j<NEW;j++)ew[j]=D(e1);
    free_hbuf(pv[0].hbuf);
#ifdef CONFIG_PTHREAD
    for(ic=1;ic<B_NUM;ic++)free_hbuf(pv[ic].hbuf);
#endif
    if(vvv)printf(" lanz2:%s",sclock(~0l));
    return(itr);}
    
/* --------------------------  Matrizen ----------------------------------- */
/*
 *  Complex funktioniert so NICHT!!!!!!!!!!!!!!!!!!!
 *
 * bei komplexen (hermitisch) Matrixen vom Typ: A+iB
 *   einfach [[ A,-B], [ B, A]] jacobieen, dann alle EWs doppelt.
 *   aber nicht grad das beste vom Ei -> eigenvektoren
 * Driver for routine jacobi fuer symmetrische reelle Matrizen
 * sparc-Ultra1-Boson: time/s = 1.5e-6 dim ^ 3.2
 * --------------------------  Matrizen ----------------------------------- */
#define ROTAT2(a,i,j,k,l) g=a[i*n+j];h=a[k*n+l];a[i*n+j]=g-s*(h+g*tau);\
	a[k*n+l]=h+s*(g-h*tau);
void jacobi2(double a[], int n, double d[], double v[], int *nrot){
    int j,iq,ip,i;
    double theta,tau,t,s,h,g,c,*b,*z,tresh,sm;
    if(!(b=(double *)malloc((size_t) (n*S_double)))) F0("malloc");
    if(!(z=(double *)malloc((size_t) (n*S_double)))) F0("malloc");
    for(ip=0;ip<n;ip++){
     for(iq=0;iq<n;iq++)v[ip*n+iq]=0.0;v[ip*n+ip]=1.0;}
    for(ip=0;ip<n;ip++){b[ip]=d[ip]=a[ip*n+ip];z[ip]=0.0;}
    *nrot=0;
    for (i=1;i<=50;i++){
      sm=0.0;
      for(ip=0;ip<n-1;ip++)for(iq=ip+1;iq<n;iq++)sm+=fabs(a[ip*n+iq]);
      if(sm==0.0){ break; } /* finished */
      if (i < 4)tresh=0.2*sm/(n*n);
	else    tresh=0.0;
      for (ip=0;ip<n-1;ip++) {
	for (iq=ip+1;iq<n;iq++) {
	 g=100.0*fabs(a[ip*n+iq]);
	 if (i>4 && (fabs(d[ip])+g)==fabs(d[ip])
		 && (fabs(d[iq])+g)==fabs(d[iq])) a[ip*n+iq]=Z(0.0);
	 else if (fabs(a[ip*n+iq]) > tresh) {
	  h=d[iq]-d[ip];
	  if ((fabs(h)+g)==fabs(h)) t=(a[ip*n+iq])/h;
	  else { theta=0.5*h/(a[ip*n+iq]);
		 t=1.0/(fabs(theta)+sqrt(1.0+theta*theta));
		 if (theta < 0.0) t = -t;}
	  c=1.0/sqrt(1+t*t);s=t*c;tau=s/(1.0+c);h=t*a[ip*n+iq];
	  z[ip] -= h;z[iq] += h;d[ip] -= h;d[iq] += h;a[ip*n+iq]=Z(0.0);
	  for (j=0;j<=ip-1;j++)    { ROTAT2(a,j,ip,j,iq)}
	  for (j=ip+1;j<=iq-1;j++) { ROTAT2(a,ip,j,j,iq)}
	  for (j=iq+1;j<n;j++)     { ROTAT2(a,ip,j,iq,j)}
	  for (j=0;j<n;j++)        { ROTAT2(v,j,ip,j,iq)}
	  ++(*nrot);
	 }
	}
      }
      for (ip=0;ip<n;ip++){ b[ip]+=z[ip]; d[ip]=b[ip]; z[ip]=0.0;}
    } 
    if(i>50) E0("Too many iterations in routine jacobi");
    free(z); free(b);
    return;
}
#undef ROTAT2
/* ----------------------------------- ende jacobi ---------------------- */
